import java.util.Vector;

/** Main parser program for simple query plans.

    Query plan format:
    
        {operator} ::-  join({operator}.{attrib} {operator}.{attrib})   (join)
                   ::-  {table}                                         (scan)
                   
 */
 
class PlanParser {
    public PlanParser() {
    }
    
    /** Finds the next non-whitespace character in a string, given a string and
        a position */
    private int nextNonWS(String query, int i) {
        while (i < query.length() && (query.charAt(i) == ' ' || query.charAt(i) == '\t'
            || query.charAt(i) == '\n' || query.charAt(i) == '\r'))
            i++;
            
        return i;
    }
    
    /** Finds the next whitespace character in a string, given a string and
        a position */
    private int nextWS(String query, int i) {
        while (i < query.length() && query.charAt(i) != ' ' && query.charAt(i) != '\t'
            && query.charAt(i) != '\n' && query.charAt(i) != '\r')
            i++;
            
        return i;
    }
    
    /** Given a position of a left paren, finds the matching right 
        paren and returns the index */
    private int matchParen(String query, int i) {
        int parens = 0;
        for (; i < query.length(); i++) {
            if (query.charAt(i) == '(')
                parens++;
            else if (query.charAt(i) == ')')
                parens--;
                
            if (parens == 0)
                return i;
        }
        return -1;
    }
    
    /** Returns true if the string begins with "join" or "join(" */
    private boolean isJoin(String query, int i) {
        if (i + 5 >= query.length())
            return false;

        return (query.substring(i, i + 4).equals("join") && (query.charAt(i + 4) == '(')
            || query.charAt(i + 4) == ' ' || query.charAt(i + 4) == '\t');
    }

    /** Attempts to split the string into left and right arguments and sub-attributes.
    
        For example, the arguments: join(x.a a.b).c d.e
        Should become:
            leftArg = join(x.a a.b)
            leftAttr = c
            rightArg = d
            rightAttr = e
            
        These are returned in the Vector
        */
    private Vector splitArgs(String args) {
        String leftAttr = null, leftArg = null, rightAttr = null, rightArg = null;
        int i = 0, j = 0;
        
        // Find the end of the left argument
        while (i < args.length()) {
            // If we found a left paren, jump to the matching right paren
            if (args.charAt(i) == '(') {
                i = matchParen(args, i) + 1;
                
            // We found the '.' which delimits the attribute name
            } else if (args.charAt(i) == '.') {
                leftArg = args.substring(0, i);
                j = nextWS(args, i);
                leftAttr = args.substring(i + 1, j);
                j++;
                break;
                
            // Some character we'll ignore    
            } else {
                i++;
            }
        }
        
        // Now we know that j is the first character of the right-hand side
        // Find the end of the right argument and verify we have "correct" syntax
        i = j;
        while (i < args.length()) {
            // If we found a left paren, jump to the matching right paren
            if (args.charAt(i) == '(') {
                i = matchParen(args, i) + 1;
                
            // We found the '.' which delimits the attribute name
            } else if (args.charAt(i) == '.') {
                rightArg = args.substring(j, i);
                j = nextWS(args, i);

                rightAttr = args.substring(i + 1, j);
                j++;
                break;
                
            // Some character we'll ignore    
            } else {
                i++;
            }
        }
        // Be sure there isn't any other non-whitespace "junk" after the end
        // of the arguments
        if (j < args.length() && nextNonWS(args, j) > -1)
            return new Vector();
            
        Vector v = new Vector();
        v.setSize(4);
        
        v.setElementAt(leftArg, 0);
        v.setElementAt(leftAttr, 1);
        v.setElementAt(rightArg, 2);
        v.setElementAt(rightAttr, 3);

        return v;
    }

    /** Parses a query plan given in our simple plan language */
    public Operator parse(String query) {
        char ch;
        int i, j;
        boolean again = true;

        // First thing should be "join" or the name
        // of a relation to scan
        if (!isJoin(query, 0)) {
            // Should be the name of a relation to scan
            
            return new Scan(query);
        }
        
        i = nextNonWS(query, 4);
        
        // Should be a paren after "join" for the arguments
        if (!(query.charAt(i) == '('))
            return null;
        
        j = matchParen(query, i);
        i++;
        
        // From i to j - 1 = the arguments of the join.  Now we need
        // to separate into left and right child
        String args = query.substring(i, j);
        
        // The join syntax is:
        //   left_child.join_attr right_child.join_attr
        // where each child might be another join
        // We need to figure out where to break things now:
        
        
        //splitArgs(args, leftArg, leftAttr, rightArg, rightAttr);
        Vector v = splitArgs(args);
        
        // Was there an error?
        if (v.isEmpty())
            return null;
        
        String leftArg = (String)v.elementAt(0);
        String leftAttr = (String)v.elementAt(1);
        String rightArg = (String)v.elementAt(2);
        String rightAttr = (String)v.elementAt(3);
        
        // For debug purposes, we may want to see what we have:
        //System.out.println(args); 
        //System.out.println(leftArg + "\t" + leftAttr); 
        //System.out.println(rightArg + "\t" + rightAttr); 
        
        // We now point at the beginning of the arguments of the join
        // We need to separate it into two children

        Operator left = parse(leftArg);
        Operator right = parse(rightArg);
        
	// Modify this to create your join operator as appropriate
	return null;
	//        return new HashJoin(left, leftAttr, right, rightAttr);
        
    }    
};
